#include <stdio.h>
#include <stdlib.h>

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "lib.h"
#include "type.h"
#include "asm_symbol.h"
#include "asm_code.h"
#include "asm_code_arm.h"

static int asm_code_arm_get_word_size(void)
{
  return 4;
}

static int asm_code_arm_get_pointer_size(void)
{
  return asm_code_arm_get_word_size();
}

static int asm_code_arm_funccall_args_reg_number(void)
{
  return 4; /* r0/r1/r2/r3 (See funccall_regs[]) */
}

static int asm_code_arm_funccall_args_stack_number(void)
{
  return VALUE_FUNCTION_ARGS_NUM;
}

static int asm_code_arm_tmp_reg_number(void)
{
  return 3; /* r4/r5/r6 (See tmp_regs[]) */
}

static int asm_code_arm_tmp_stack_number(void)
{
  return VALUE_FUNCTION_ARGS_NUM;
}

static int asm_code_arm_function_register_number(void)
{
  return 4; /* r4/r5/r6/r7 (See asm_code_arm_function_register_save() */
}

static int asm_code_arm_function_saveparam_number(void)
{
  return 4; /* fp/ip/lr/pc (See asm_code_arm_function_start() */
}

static int asm_code_arm_stack_align_size(void)
{
  return 16;
}

static int asm_code_arm_stack_correct_size(void)
{
  return 0;
}

static char *load_by_size(model_t model)
{
  if (model != NULL) {
    switch (model_get_size(model)) {
    case -1: ERREXIT(1);
    case  1: return "ldrb";
    case  2: return "ldrh";
    default: break;
    }
  }
  return "ldr";
}

static char *store_by_size(model_t model)
{
  if (model != NULL) {
    switch (model_get_size(model)) {
    case -1: ERREXIT(1);
    case  1: return "strb";
    case  2: return "strh";
    default: break;
    }
  }
  return "str";
}

static void asm_code_arm_get_value(FILE *out, long value)
{
  ASM_CODE_OUT(out, "\tmov\tr0, #0x%lx\n", value & 0xff000000);
  ASM_CODE_OUT(out, "\tadd\tr0, #0x%lx\n", value & 0xff0000);
  ASM_CODE_OUT(out, "\tadd\tr0, #0x%lx\n", value & 0xff00);
  ASM_CODE_OUT(out, "\tadd\tr0, #0x%lx\n", value & 0xff);
}

static void asm_code_arm_get_value_r1(FILE *out, long value)
{
  ASM_CODE_OUT(out, "\tmov\tr1, #0x%lx\n", value & 0xff000000);
  ASM_CODE_OUT(out, "\tadd\tr1, #0x%lx\n", value & 0xff0000);
  ASM_CODE_OUT(out, "\tadd\tr1, #0x%lx\n", value & 0xff00);
  ASM_CODE_OUT(out, "\tadd\tr1, #0x%lx\n", value & 0xff);
}

static void asm_code_arm_get_address_stack(FILE *out)
{
  ASM_CODE_OUT(out, "\tmov\tr0, sp\n");
}

static void asm_code_arm_get_address(FILE *out, char *label)
{
  ASM_CODE_OUT(out, "\tldr\tr0, [pc, #0]\n");
  ASM_CODE_OUT(out, "\tmov\tpc, pc\n");
  ASM_CODE_OUT(out, "\t.word\t%s\n", label);
}

static void asm_code_arm_add_address(FILE *out, int offset)
{
  ASM_CODE_OUT(out, "\tadd\tr0, r0, #0x%x\n", offset);
}

static void asm_code_arm_get_r1(FILE *out)
{
  ASM_CODE_OUT(out, "\tmov\tr0, r1\n");
}

static void asm_code_arm_set_r1(FILE *out)
{
  ASM_CODE_OUT(out, "\tmov\tr1, r0\n");
}

static void asm_code_arm_memory_load(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\t%s\tr0, [r1, #0x%x]\n", load_by_size(model), offset);
}

static void asm_code_arm_memory_store(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\t%s\tr0, [r1, #0x%x]\n", store_by_size(model), offset);
}

static void asm_code_arm_stack_load(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\t%s\tr0, [sp, #0x%x]\n", load_by_size(model), offset);
}

static void asm_code_arm_stack_store(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\t%s\tr0, [sp, #0x%x]\n", store_by_size(model), offset);
}

static void asm_code_arm_stack_load_r1(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\t%s\tr1, [sp, #0x%x]\n", load_by_size(model), offset);
}

static void asm_code_arm_stack_store_r1(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\t%s\tr1, [sp, #0x%x]\n", store_by_size(model), offset);
}

static void asm_code_arm_stack_expand(FILE *out, int size)
{
  ASM_CODE_OUT(out, "\tsub\tsp, sp, #0x%x\n", size);
}

static void asm_code_arm_stack_reduce(FILE *out, int size)
{
  ASM_CODE_OUT(out, "\tadd\tsp, sp, #0x%x\n", size);
}

static char *funccall_regs[] = {
  "r0", "r1", "r2", "r3"
};

static void asm_code_arm_funccall_reg_load(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmov\tr0, %s\n", funccall_regs[number]);
}

static void asm_code_arm_funccall_reg_store(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmov\t%s, r0\n", funccall_regs[number]);
}

static char *tmp_regs[] = {
  "r4", "r5", "r6",
};

static void asm_code_arm_tmp_reg_load(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmov\tr0, %s\n", tmp_regs[number]);
}

static void asm_code_arm_tmp_reg_save(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmov\t%s, r0\n", tmp_regs[number]);
}

static void asm_code_arm_tmp_reg_load_r1(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmov\tr1, %s\n", tmp_regs[number]);
}

static void asm_code_arm_tmp_reg_save_r1(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmov\t%s, r1\n", tmp_regs[number]);
}

static int asm_code_arm_sign_extension_char(FILE *out)
{
  ASM_CODE_OUT(out, "\tlsl\tr0, r0, #24\n");
  ASM_CODE_OUT(out, "\tasr\tr0, r0, #24\n");
  return 0;
}

static int asm_code_arm_sign_extension_uchar(FILE *out)
{
  ASM_CODE_OUT(out, "\tand\tr0, r0, #0xff\n");
  return 0;
}

static int asm_code_arm_sign_extension_short(FILE *out)
{
  ASM_CODE_OUT(out, "\tlsl\tr0, r0, #16\n");
  ASM_CODE_OUT(out, "\tasr\tr0, r0, #16\n");
  return 0;
}

static int asm_code_arm_sign_extension_ushort(FILE *out)
{
  ASM_CODE_OUT(out, "\tlsl\tr0, r0, #16\n");
  ASM_CODE_OUT(out, "\tlsr\tr0, r0, #16\n");
  return 0;
}

static int asm_code_arm_sign_extension_int(FILE *out)
{
  return -1;
}

static int asm_code_arm_sign_extension_uint(FILE *out)
{
  return -1;
}

static int asm_code_arm_calc_inv(FILE *out, model_t model, model_t model_arg0)
{
  ASM_CODE_OUT(out, "\tmvn\tr0, r0\n");
  return 0;
}

static int asm_code_arm_calc_minus(FILE *out, model_t model, model_t model_arg0)
{
  ASM_CODE_OUT(out, "\trsb\tr0, r0, #0\n");
  return 0;
}

static int asm_code_arm_calc_op1(FILE *out, c_type_t type, model_t model, model_t model_arg0)
{
  return -1;
}

static void asm_code_arm_calc_add(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tadd\tr0, r0, r1\n");
}

static void asm_code_arm_calc_sub(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\trsb\tr0, r1, r0\n");
}

static void asm_code_arm_calc_and(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tand\tr0, r0, r1\n");
}

static void asm_code_arm_calc_or(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\torr\tr0, r0, r1\n");
}

static void asm_code_arm_calc_xor(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\teor\tr0, r0, r1\n");
}

static int asm_code_arm_calc_mul(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tmul\tr0, r0, r1\n");
  return 0;
}

static int asm_code_arm_calc_div(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  return -1; /* use builtin library */
}

static int asm_code_arm_calc_mod(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  return -1; /* use builtin library */
}

static int asm_code_arm_calc_llshift(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tlsl\tr0, r0, r1\n");
  return 0;
}

static int asm_code_arm_calc_rashift(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tasr\tr0, r0, r1\n");
  return 0;
}

static int asm_code_arm_calc_rlshift(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tlsr\tr0, r0, r1\n");
  return 0;
}

static int asm_code_arm_calc_op2(FILE *out, c_type_t type, model_t model, model_t model_arg0, model_t model_arg1)
{
  return -1;
}

static void asm_code_arm_branch(FILE *out, char *label)
{
  ASM_CODE_OUT(out, "\tb\t%s\n", label);
}

static int asm_code_arm_branch_zero(FILE *out, char *label)
{
  ASM_CODE_OUT(out, "\tcmp\tr0, #0\n");
  ASM_CODE_OUT(out, "\tbeq\t%s\n", label);
  return 0;
}

static int asm_code_arm_branch_nzero(FILE *out, char *label)
{
  ASM_CODE_OUT(out, "\tcmp\tr0, #0\n");
  ASM_CODE_OUT(out, "\tbne\t%s\n", label);
  return 0;
}

static int asm_code_arm_branch_cmp_eq(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\tr0, r1\n");
  ASM_CODE_OUT(out, "\tbeq\t%s\n", label);
  return 0;
}

static int asm_code_arm_branch_cmp_ne(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\tr0, r1\n");
  ASM_CODE_OUT(out, "\tbne\t%s\n", label);
  return 0;
}

static int asm_code_arm_branch_cmp_lt(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\tr0, r1\n");
  ASM_CODE_OUT(out, "\tblt\t%s\n", label);
  return 0;
}

static int asm_code_arm_branch_cmp_gt(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\tr0, r1\n");
  ASM_CODE_OUT(out, "\tbgt\t%s\n", label);
  return 0;
}

static int asm_code_arm_branch_cmp_le(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\tr0, r1\n");
  ASM_CODE_OUT(out, "\tble\t%s\n", label);
  return 0;
}

static int asm_code_arm_branch_cmp_ge(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\tr0, r1\n");
  ASM_CODE_OUT(out, "\tbge\t%s\n", label);
  return 0;
}

static int asm_code_arm_branch_cmp_ult(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\tr0, r1\n");
  ASM_CODE_OUT(out, "\tbcc\t%s\n", label);
  return 0;
}

static int asm_code_arm_branch_cmp_ugt(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\tr0, r1\n");
  ASM_CODE_OUT(out, "\tbhi\t%s\n", label);
  return 0;
}

static int asm_code_arm_branch_cmp_ule(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\tr0, r1\n");
  ASM_CODE_OUT(out, "\tbls\t%s\n", label);
  return 0;
}

static int asm_code_arm_branch_cmp_uge(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\tr0, r1\n");
  ASM_CODE_OUT(out, "\tbcs\t%s\n", label);
  return 0;
}

static int asm_code_arm_branch_cmp(FILE *out, char *label, c_type_t type, model_t model)
{
  return -1;
}

static int asm_code_arm_function_call(FILE *out, char *label)
{
  ASM_CODE_OUT(out, "\tbl\t%s\n", label);
  return 0;
}

static int asm_code_arm_function_call_set(FILE *out)
{
  ASM_CODE_OUT(out, "\tmov\tr7, r0\n");
  return 0;
}

static int asm_code_arm_function_call_pointer(FILE *out)
{
  ASM_CODE_OUT(out, "\tmov\tlr, pc\n");
  ASM_CODE_OUT(out, "\tmov\tpc, r7\n");
  return 0;
}

static void asm_code_arm_function_start(FILE *out)
{
  ASM_CODE_OUT(out, "\tmov\tip, sp\n");
  ASM_CODE_OUT(out, "\tpush\t{fp, ip, lr, pc}\n");
}

static void asm_code_arm_function_end(FILE *out)
{
  ASM_CODE_OUT(out, "\tldm\tsp, {fp, sp, pc}\n");
}

static void asm_code_arm_function_register_save(FILE *out)
{
  ASM_CODE_OUT(out, "\tpush\t{r4, r5, r6, r7}\n");
}

static void asm_code_arm_function_register_load(FILE *out)
{
  ASM_CODE_OUT(out, "\tpop\t{r4, r5, r6, r7}\n");
}

struct asm_code_callback asm_code_callback_arm = {
  asm_code_arm_get_word_size,
  asm_code_arm_get_pointer_size,

  asm_code_arm_funccall_args_reg_number,
  asm_code_arm_funccall_args_stack_number,
  asm_code_arm_tmp_reg_number,
  asm_code_arm_tmp_stack_number,
  asm_code_arm_function_register_number,
  asm_code_arm_function_saveparam_number,

  asm_code_arm_stack_align_size,
  asm_code_arm_stack_correct_size,

  asm_code_arm_get_value,
  asm_code_arm_get_value_r1,
  asm_code_arm_get_address_stack,
  asm_code_arm_get_address,
  asm_code_arm_add_address,
  asm_code_arm_get_r1,
  asm_code_arm_set_r1,

  asm_code_arm_memory_load,
  asm_code_arm_memory_store,
  asm_code_arm_stack_load,
  asm_code_arm_stack_store,
  asm_code_arm_stack_load_r1,
  asm_code_arm_stack_store_r1,
  asm_code_arm_stack_expand,
  asm_code_arm_stack_reduce,

  asm_code_arm_funccall_reg_load,
  asm_code_arm_funccall_reg_store,
  asm_code_arm_tmp_reg_load,
  asm_code_arm_tmp_reg_save,
  asm_code_arm_tmp_reg_load_r1,
  asm_code_arm_tmp_reg_save_r1,

  asm_code_arm_sign_extension_char,
  asm_code_arm_sign_extension_uchar,
  asm_code_arm_sign_extension_short,
  asm_code_arm_sign_extension_ushort,
  asm_code_arm_sign_extension_int,
  asm_code_arm_sign_extension_uint,

  asm_code_arm_calc_inv,
  asm_code_arm_calc_minus,

  asm_code_arm_calc_op1,

  asm_code_arm_calc_add,
  asm_code_arm_calc_sub,
  asm_code_arm_calc_and,
  asm_code_arm_calc_or,
  asm_code_arm_calc_xor,

  asm_code_arm_calc_mul,
  asm_code_arm_calc_div,
  asm_code_arm_calc_mod,
  asm_code_arm_calc_llshift,
  asm_code_arm_calc_rashift,
  asm_code_arm_calc_rlshift,

  asm_code_arm_calc_op2,

  asm_code_arm_branch,
  asm_code_arm_branch_zero,
  asm_code_arm_branch_nzero,

  asm_code_arm_branch_cmp_eq,
  asm_code_arm_branch_cmp_ne,

  asm_code_arm_branch_cmp_lt,
  asm_code_arm_branch_cmp_gt,
  asm_code_arm_branch_cmp_le,
  asm_code_arm_branch_cmp_ge,
  asm_code_arm_branch_cmp_ult,
  asm_code_arm_branch_cmp_ugt,
  asm_code_arm_branch_cmp_ule,
  asm_code_arm_branch_cmp_uge,

  asm_code_arm_branch_cmp,

  asm_code_arm_function_call,
  asm_code_arm_function_call_set,
  asm_code_arm_function_call_pointer,

  asm_code_arm_function_start,
  asm_code_arm_function_end,
  asm_code_arm_function_register_save,
  asm_code_arm_function_register_load,
};
